/******************************************************************************
*                                                                             *
*    diblib.c                               Copyright(c) 2004-2007 itow,y.    *
*                                                                             *
******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include "diblib.h"


/*
#define DIB_ROW_BYTES(width,bpp) (((width)*(bpp)+31)/32*4)
*/
#define DIB_ROW_BYTES(width,bpp) ((((width)*(bpp)+31)>>5)<<2)




BOOL DIB_Create(DIBInfo **ppInfo,int nWidth,int nHeight,int nBitCount)
{
	*ppInfo=(DIBInfo*)malloc(sizeof(DIBInfo));
	if (*ppInfo==NULL)
		return FALSE;
	(*ppInfo)->pbmi=(BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER)+
							(nBitCount<=8?(1<<nBitCount)*sizeof(RGBQUAD):0));
	if ((*ppInfo)->pbmi==NULL) {
		free(*ppInfo);
		*ppInfo=NULL;
		return FALSE;
	}
	(*ppInfo)->pbmi->bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	(*ppInfo)->pbmi->bmiHeader.biWidth=nWidth;
	(*ppInfo)->pbmi->bmiHeader.biHeight=nHeight;
	(*ppInfo)->pbmi->bmiHeader.biPlanes=1;
	(*ppInfo)->pbmi->bmiHeader.biBitCount=nBitCount;
	(*ppInfo)->pbmi->bmiHeader.biCompression=BI_RGB;
	(*ppInfo)->pbmi->bmiHeader.biSizeImage=0;
	(*ppInfo)->pbmi->bmiHeader.biXPelsPerMeter=0;
	(*ppInfo)->pbmi->bmiHeader.biYPelsPerMeter=0;
	(*ppInfo)->pbmi->bmiHeader.biClrUsed=0;
	(*ppInfo)->pbmi->bmiHeader.biClrImportant=0;
	(*ppInfo)->hbm=CreateDIBSection(NULL,(*ppInfo)->pbmi,DIB_RGB_COLORS,
													&(*ppInfo)->pBits,NULL,0);
	if ((*ppInfo)->hbm==NULL) {
		free((*ppInfo)->pbmi);
		free(*ppInfo);
		*ppInfo=NULL;
		return FALSE;
	}
	(*ppInfo)->nWidth=nWidth;
	(*ppInfo)->nHeight=nHeight;
	(*ppInfo)->nBitsPerPixel=nBitCount;
	(*ppInfo)->hdcMem=NULL;
	return TRUE;
}


BOOL DIB_Load(DIBInfo **ppInfo,LPCTSTR pszFileName)
{
	HANDLE hFile;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	DWORD dwSize,dwRead;

	hFile=CreateFile(pszFileName,GENERIC_READ,FILE_SHARE_READ,NULL,
									OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if (hFile==INVALID_HANDLE_VALUE)
		return FALSE;
	if (!ReadFile(hFile,&bmfh,sizeof(BITMAPFILEHEADER),&dwRead,NULL)
										|| sizeof(BITMAPFILEHEADER)!=dwRead
										|| bmfh.bfType!=0x4D42) {
		CloseHandle(hFile);
		return FALSE;
	}
	if (!ReadFile(hFile,&bmih,sizeof(BITMAPINFOHEADER),&dwRead,NULL)
									|| sizeof(BITMAPINFOHEADER)!=dwRead
									|| bmih.biSize!=sizeof(BITMAPINFOHEADER)
									|| bmih.biHeight<=0
									|| bmih.biCompression!=BI_RGB) {
		CloseHandle(hFile);
		return FALSE;
	}
	if (!DIB_Create(ppInfo,bmih.biWidth,bmih.biHeight,bmih.biBitCount)) {
		CloseHandle(hFile);
		return FALSE;
	}
	if (bmih.biBitCount<=8) {
		dwSize=(1<<bmih.biBitCount)*sizeof(RGBQUAD);
		if (!ReadFile(hFile,(*ppInfo)->pbmi->bmiColors,dwSize,&dwRead,NULL)
														|| dwRead!=dwSize) {
			DIB_Destroy(ppInfo);
			CloseHandle(hFile);
			return FALSE;
		}
	}
	dwSize=DIB_ROW_BYTES(bmih.biWidth,bmih.biBitCount)*bmih.biHeight;
	if (!ReadFile(hFile,(*ppInfo)->pBits,dwSize,&dwRead,NULL)
														|| dwRead!=dwSize) {
		DIB_Destroy(ppInfo);
		CloseHandle(hFile);
		return FALSE;
	}
	CloseHandle(hFile);
	return TRUE;
}


void DIB_Destroy(DIBInfo **ppInfo)
{
	if (*ppInfo!=NULL) {
		if ((*ppInfo)->hbm!=NULL)
			DeleteObject((*ppInfo)->hbm);
		if ((*ppInfo)->pbmi!=NULL)
			free((*ppInfo)->pbmi);
		if ((*ppInfo)->hdcMem!=NULL)
			DeleteDC((*ppInfo)->hdcMem);
		free(*ppInfo);
		*ppInfo=NULL;
	}
}


static void NormalCopy(DIBInfo *pDstInfo,int nDstLeft,int nDstTop,int nWidth,
				int nHeight,const DIBInfo *pSrcInfo,int nSrcLeft,int nSrcTop)
{
	const BYTE *p;
	BYTE *q;
	unsigned int uSrcRowBytes,uDstRowBytes;
	int x,y,i;
	const RGBQUAD *prgb;

	uSrcRowBytes=DIB_ROW_BYTES(pSrcInfo->nWidth,pSrcInfo->nBitsPerPixel);
	uDstRowBytes=DIB_ROW_BYTES(pDstInfo->nWidth,pDstInfo->nBitsPerPixel);
	if (pDstInfo->nBitsPerPixel==pSrcInfo->nBitsPerPixel) {
		p=(BYTE*)pSrcInfo->pBits+(pSrcInfo->nHeight-1-nSrcTop)*uSrcRowBytes+
											nSrcLeft*pSrcInfo->nBitsPerPixel/8;
		q=(BYTE*)pDstInfo->pBits+(pDstInfo->nHeight-1-nDstTop)*uDstRowBytes+
											nDstLeft*pDstInfo->nBitsPerPixel/8;
		for (y=0;y<nHeight;y++) {
			memcpy(q,p,nWidth);
			p-=uSrcRowBytes;
			q-=uDstRowBytes;
		}
		return;
	}
	switch (pDstInfo->nBitsPerPixel) {
	case 24:
		switch (pSrcInfo->nBitsPerPixel) {
		case 1:
			{
				int Shift;

				prgb=pSrcInfo->pbmi->bmiColors;
				for (y=0;y<nHeight;y++) {
					p=(BYTE*)pSrcInfo->pBits+
								(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes;
					q=(BYTE*)(BYTE*)pDstInfo->pBits+
						(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
					Shift=7-nSrcLeft%8;
					for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
						i=(*p>>Shift)&0x01;
						*q++=prgb[i].rgbBlue;
						*q++=prgb[i].rgbGreen;
						*q++=prgb[i].rgbRed;
						if (Shift==0) {
							Shift=7;
							p++;
						} else
							Shift--;
					}
				}
			}
			break;
		case 4:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
					(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft/2;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
					i=((x&1)==0)?(*p>>4):(*p++&0x0F);
					*q++=prgb[i].rgbBlue;
					*q++=prgb[i].rgbGreen;
					*q++=prgb[i].rgbRed;
				}
			}
			break;
		case 8:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
						(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=0;x<nWidth;x++) {
					i=*p++;
					*q++=prgb[i].rgbBlue;
					*q++=prgb[i].rgbGreen;
					*q++=prgb[i].rgbRed;
				}
			}
			break;
		}
		break;
	}
}


static void NormalAlphaCopy(DIBInfo *pDstInfo,int nDstLeft,int nDstTop,
	int nWidth,int nHeight,const DIBInfo *pSrcInfo,int nSrcLeft,int nSrcTop,
																int nOpacity)
{
	const BYTE *p;
	BYTE *q;
	unsigned int uSrcRowBytes,uDstRowBytes;
	int nAlpha;
	int x,y,i;
	const RGBQUAD *prgb;

	uSrcRowBytes=DIB_ROW_BYTES(pSrcInfo->nWidth,pSrcInfo->nBitsPerPixel);
	uDstRowBytes=DIB_ROW_BYTES(pDstInfo->nWidth,pDstInfo->nBitsPerPixel);
	nAlpha=256-nOpacity;
	switch (pDstInfo->nBitsPerPixel) {
	case 24:
		switch (pSrcInfo->nBitsPerPixel) {
		case 1:
			{
				int Shift;

				prgb=pSrcInfo->pbmi->bmiColors;
				for (y=0;y<nHeight;y++) {
					p=(BYTE*)pSrcInfo->pBits+
								(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes;
					q=(BYTE*)pDstInfo->pBits+
						(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
					Shift=7-nSrcLeft%8;
					for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
						i=(*p>>Shift)&0x01;
						*q=((*q)*nAlpha+prgb[i].rgbBlue*nOpacity)>>8;
						q++;
						*q=((*q)*nAlpha+prgb[i].rgbGreen*nOpacity)>>8;
						q++;
						*q=((*q)*nAlpha+prgb[i].rgbRed*nOpacity)>>8;
						q++;
						if (Shift==0) {
							Shift=7;
							p++;
						} else
							Shift--;
					}
				}
			}
			break;
		case 4:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
					(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft/2;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
					i=((x&1)==0)?(*p>>4):(*p++&0x0F);
					*q=((*q)*nAlpha+prgb[i].rgbBlue*nOpacity)>>8;
					q++;
					*q=((*q)*nAlpha+prgb[i].rgbGreen*nOpacity)>>8;
					q++;
					*q=((*q)*nAlpha+prgb[i].rgbRed*nOpacity)>>8;
					q++;
				}
			}
			break;
		case 8:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
						(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=0;x<nWidth;x++) {
					i=*p++;
					*q=((*q)*nAlpha+prgb[i].rgbBlue*nOpacity)>>8;
					q++;
					*q=((*q)*nAlpha+prgb[i].rgbGreen*nOpacity)>>8;
					q++;
					*q=((*q)*nAlpha+prgb[i].rgbRed*nOpacity)>>8;
					q++;
				}
			}
			break;
		case 24:
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
					(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft*3;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=0;x<nWidth;x++) {
					*q=((*q)*nAlpha+(*p++)*nOpacity)>>8;
					q++;
					*q=((*q)*nAlpha+(*p++)*nOpacity)>>8;
					q++;
					*q=((*q)*nAlpha+(*p++)*nOpacity)>>8;
					q++;
				}
			}
			break;
		}
		break;
	}
}


static void TransparentCopy(DIBInfo *pDstInfo,int nDstLeft,int nDstTop,
	int nWidth,int nHeight,const DIBInfo *pSrcInfo,int nSrcLeft,int nSrcTop)
{
	const BYTE *p;
	BYTE *q;
	unsigned int uSrcRowBytes,uDstRowBytes;
	int x,y,i;
	const RGBQUAD *prgb;

	uSrcRowBytes=DIB_ROW_BYTES(pSrcInfo->nWidth,pSrcInfo->nBitsPerPixel);
	uDstRowBytes=DIB_ROW_BYTES(pDstInfo->nWidth,pDstInfo->nBitsPerPixel);
	switch (pDstInfo->nBitsPerPixel) {
	case 8:
		switch (pSrcInfo->nBitsPerPixel) {
		case 8:
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
						(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft;
				q=(BYTE*)pDstInfo->pBits+
						(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft;
				for (x=0;x<nWidth;x++) {
					if (*p!=0)
						*q=*p;
					q++;
					p++;
				}
			}
			break;
		}
		break;
	case 24:
		switch (pSrcInfo->nBitsPerPixel) {
		case 1:
			{
				int Shift;

				prgb=&pSrcInfo->pbmi->bmiColors[1];
				for (y=0;y<nHeight;y++) {
					p=(BYTE*)pSrcInfo->pBits+
								(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes;
					q=(BYTE*)pDstInfo->pBits+
						(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
					Shift=7-nSrcLeft%8;
					for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
						if (((*p>>Shift)&0x01)!=0) {
							q[0]=prgb->rgbBlue;
							q[1]=prgb->rgbGreen;
							q[2]=prgb->rgbRed;
						}
						q+=3;
						if (Shift==0) {
							Shift=7;
							p++;
						} else
							Shift--;
					}
				}
			}
			break;
		case 4:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
					(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft/2;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
					i=((x&1)==0)?(*p>>4):(*p++&0x0F);
					if (i!=0) {
						q[0]=prgb[i].rgbBlue;
						q[1]=prgb[i].rgbGreen;
						q[2]=prgb[i].rgbRed;
					}
					q+=3;
				}
			}
			break;
		case 8:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
						(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=0;x<nWidth;x++) {
					i=*p++;
					if (i!=0) {
						q[0]=prgb[i].rgbBlue;
						q[1]=prgb[i].rgbGreen;
						q[2]=prgb[i].rgbRed;
					}
					q+=3;
				}
			}
			break;
		case 24:
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
					(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft*3;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=0;x<nWidth;x++) {
					if (p[0]!=0 || p[1]!=255 || p[2]!=0) {
						q[0]=p[0];
						q[1]=p[1];
						q[2]=p[2];
					}
					q+=3;
					p+=3;
				}
			}
			break;
		}
		break;
	}
}


static void TransparentAlphaCopy(DIBInfo *pDstInfo,int nDstLeft,int nDstTop,
	int nWidth,int nHeight,const DIBInfo *pSrcInfo,int nSrcLeft,int nSrcTop,
																int nOpacity)
{
	const BYTE *p;
	BYTE *q;
	unsigned int uSrcRowBytes,uDstRowBytes;
	int nAlpha;
	int x,y,i;
	const RGBQUAD *prgb;

	uSrcRowBytes=DIB_ROW_BYTES(pSrcInfo->nWidth,pSrcInfo->nBitsPerPixel);
	uDstRowBytes=DIB_ROW_BYTES(pDstInfo->nWidth,pDstInfo->nBitsPerPixel);
	nAlpha=256-nOpacity;
	switch (pDstInfo->nBitsPerPixel) {
	case 24:
		switch (pSrcInfo->nBitsPerPixel) {
		case 1:
			{
				int Shift;

				prgb=&pSrcInfo->pbmi->bmiColors[1];
				for (y=0;y<nHeight;y++) {
					p=(BYTE*)pSrcInfo->pBits+
								(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes;
					q=(BYTE*)pDstInfo->pBits+
						(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
					Shift=7-nSrcLeft%8;
					for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
						if (((*p>>Shift)&0x01)!=0) {
							q[0]=(prgb->rgbBlue*nOpacity+q[0]*nAlpha)>>8;
							q[1]=(prgb->rgbGreen*nOpacity+q[1]*nAlpha)>>8;
							q[2]=(prgb->rgbRed*nOpacity+q[2]*nAlpha)>>8;
						}
						q+=3;
						if (Shift==0) {
							Shift=7;
							p++;
						} else
							Shift--;
					}
				}
			}
			break;
		case 4:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
					(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft/2;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=nSrcLeft;x<nSrcLeft+nWidth;x++) {
					i=((x&1)==0)?(*p>>4):(*p++&0x0F);
					if (i!=0) {
						q[0]=(prgb[i].rgbBlue*nOpacity+q[0]*nAlpha)>>8;
						q[1]=(prgb[i].rgbGreen*nOpacity+q[1]*nAlpha)>>8;
						q[2]=(prgb[i].rgbRed*nOpacity+q[2]*nAlpha)>>8;
					}
					q+=3;
				}
			}
			break;
		case 8:
			prgb=pSrcInfo->pbmi->bmiColors;
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
						(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=0;x<nWidth;x++) {
					i=*p++;
					if (i!=0) {
						q[0]=(prgb[i].rgbBlue*nOpacity+q[0]*nAlpha)>>8;
						q[1]=(prgb[i].rgbGreen*nOpacity+q[1]*nAlpha)>>8;
						q[2]=(prgb[i].rgbRed*nOpacity+q[2]*nAlpha)>>8;
					}
					q+=3;
				}
			}
			break;
		case 24:
			for (y=0;y<nHeight;y++) {
				p=(BYTE*)pSrcInfo->pBits+
					(pSrcInfo->nHeight-1-y-nSrcTop)*uSrcRowBytes+nSrcLeft*3;
				q=(BYTE*)pDstInfo->pBits+
					(pDstInfo->nHeight-1-y-nDstTop)*uDstRowBytes+nDstLeft*3;
				for (x=0;x<nWidth;x++) {
					if (p[0]!=0 || p[1]!=255 || p[2]!=0) {
						q[0]=(p[0]*nOpacity+q[0]*nAlpha)>>8;
						q[1]=(p[1]*nOpacity+q[1]*nAlpha)>>8;
						q[2]=(p[2]*nOpacity+q[2]*nAlpha)>>8;
					}
					q+=3;
					p+=3;
				}
			}
			break;
		}
		break;
	}
}


void DIB_Copy(DIBInfo *pDstInfo,int nDstLeft,int nDstTop,int nWidth,
	int nHeight,const DIBInfo *pSrcInfo,int nSrcLeft,int nSrcTop,int nOpacity,
															DIBCopyMode Mode)
{
	if (nDstLeft<0) {
		nWidth+=nDstLeft;
		nDstLeft=0;
	}
	if (nDstTop<0) {
		nHeight+=nDstTop;
		nDstTop=0;
	}
	if (nDstLeft+nWidth>pDstInfo->nWidth)
		nWidth=pDstInfo->nWidth-nDstLeft;
	if (nDstTop+nHeight>pDstInfo->nHeight)
		nHeight=pDstInfo->nHeight-nDstTop;
	if (nSrcLeft+nWidth>pSrcInfo->nWidth)
		nWidth=pSrcInfo->nWidth-nSrcLeft;
	if (nSrcTop+nHeight>pSrcInfo->nHeight)
		nHeight=pSrcInfo->nHeight-nSrcTop;
	if (nWidth<=0 || nHeight<=0)
		return;
	switch (Mode) {
	case DIB_COPY_NORMAL:
		if (nOpacity>=DIB_OPACITY_OPAQUE)
			NormalCopy(pDstInfo,nDstLeft,nDstTop,nWidth,nHeight,pSrcInfo,
															nSrcLeft,nSrcTop);
		else
			NormalAlphaCopy(pDstInfo,nDstLeft,nDstTop,nWidth,nHeight,pSrcInfo,
													nSrcLeft,nSrcTop,nOpacity);
		break;
	case DIB_COPY_TRANSPARENT:
		if (nOpacity>=DIB_OPACITY_OPAQUE)
			TransparentCopy(pDstInfo,nDstLeft,nDstTop,nWidth,nHeight,pSrcInfo,
															nSrcLeft,nSrcTop);
		else
			TransparentAlphaCopy(pDstInfo,nDstLeft,nDstTop,nWidth,nHeight,
										pSrcInfo,nSrcLeft,nSrcTop,nOpacity);
		break;
	}
}


void DIB_FillRect(DIBInfo *pInfo,int nLeft,int nTop,int nWidth,int nHeight,
												COLORREF crColor,int nOpacity)
{
	int nRight,nBottom;
	BYTE *p;
	unsigned int uRowBytes;
	BYTE r,g,b;
	int x,y;

	nRight=nLeft+nWidth;
	nBottom=nTop+nHeight;
	if (nLeft<0)
		nLeft=0;
	if (nTop<0)
		nTop=0;
	if (nRight>pInfo->nWidth)
		nRight=pInfo->nWidth;
	if (nBottom>pInfo->nHeight)
		nBottom=pInfo->nHeight;
	if (nLeft>=nRight || nTop>=nBottom)
		return;
	uRowBytes=DIB_ROW_BYTES(pInfo->nWidth,pInfo->nBitsPerPixel);
	r=GetRValue(crColor);
	g=GetGValue(crColor);
	b=GetBValue(crColor);
	if (nOpacity>=256) {
		switch (pInfo->nBitsPerPixel) {
		case 24:
			for (y=nTop;y<nBottom;y++) {
				p=(BYTE*)pInfo->pBits+(pInfo->nHeight-1-y)*uRowBytes+nLeft*3;
				for (x=nLeft;x<nRight;x++) {
					*p++=b;
					*p++=g;
					*p++=r;
				}
			}
			break;
		}
	} else {
		int nAlpha=256-nOpacity;

		r*=nOpacity;
		g*=nOpacity;
		b*=nOpacity;
		switch (pInfo->nBitsPerPixel) {
		case 24:
			for (y=nTop;y<nBottom;y++) {
				p=(BYTE*)pInfo->pBits+(pInfo->nHeight-1-y)*uRowBytes+nLeft*3;
				for (x=nLeft;x<nRight;x++) {
					*p=(*p*nAlpha+b)>>8;
					p++;
					*p=(*p*nAlpha+g)>>8;
					p++;
					*p=(*p*nAlpha+r)>>8;
					p++;
				}
			}
			break;
		}
	}
}


BOOL DIB_DrawText(DIBInfo *pInfo,int nLeft,int nTop,LPCTSTR pszText,
									int nLength,HFONT hfont,COLORREF crColor)
{
	HBITMAP hbmOld;
	HFONT hfontOld;

	if (pInfo->hdcMem==NULL) {
		pInfo->hdcMem=CreateCompatibleDC(NULL);
		if (pInfo->hdcMem==NULL)
			return FALSE;
	}
	hbmOld=(HBITMAP)SelectObject(pInfo->hdcMem,pInfo->hbm);
	hfontOld=(HFONT)SelectObject(pInfo->hdcMem,hfont);
	SetBkMode(pInfo->hdcMem,TRANSPARENT);
	SetTextColor(pInfo->hdcMem,crColor);
	TextOut(pInfo->hdcMem,nLeft,nTop,pszText,nLength);
	SelectObject(pInfo->hdcMem,hbmOld);
	SelectObject(pInfo->hdcMem,hfontOld);
	return TRUE;
}
